
package fr.eurecom.ratcom.util;


import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;

import fr.eurecom.ratcom.properties.ProjectProperties;



public class Logging {

	static ProjectProperties properties = ProjectProperties.getInstance();
	/**
	 * Singleton Constructor for the default logger object. Depending on the configuration
	 * of the framework, this logger may print to the console or to the default log-file. If
	 * printed to the console, no log-file will be created.   
	 * @return The logging instance
	 */
	public static Logging getLogger() {
		if(instance == null) {
				if(properties.outputToConsole) {
					instance = new Logging(System.out);
				} else {
					return getLogger(properties.logFileName);
				}
		}
		return instance;
	}

	/**
	 * Returns an instance of the specified logger object. 
	 * The logger object is identified by its name, which is the name 
	 * of the file to which the log statements are written.
	 * <p>
	 * Upon calling this method the first time with a given logFileName, a
	 * new logger object is created. All subsequent calls with the same logFileName
	 * will return the corresponding logger object created before.
	 * <p>
	 * By default, the log-file is stored in the directory 'logs'. The logFileName 
	 * may contain sub-directories, which will be created if necessary.
	 * 
	 * If the flag logToTimeDirectory is set to true in the configuration file
	 * of the current project, the log-file is placed in the directory 'logs/XXX', where
	 * XXX is a unique name composed of the project name and the time when the current
	 * simulation started. Use this feature to not overwrite log-files in subsequent runs.
	 * @param aName The file name that identifies the logger object to be returned. 
	 * This name may also contain sub-directories, in which the corresponding log file is
	 * placed.
	 * @return A logging object for the given log file name.  
	 */
	public static Logging getLogger(String aName) {
		return getLogger(aName, false);
	}

	/**
	 * Returns an instance of the specified logger object. 
	 * The logger object is identified by its name, which is the name 
	 * of the file to which the log statements are written.
	 * <p>
	 * Upon calling this method the first time with a given logFileName, a
	 * new logger object is created. All subsequent calls with the same logFileName
	 * will return the corresponding logger object created before.
	 * <p>
	 * By default, the log-file is stored in the directory 'logs'. The logFileName 
	 * may contain sub-directories, which will be created if necessary.
	 * 
	 * If the flag logToTimeDirectory is set to true in the configuration file
	 * of the current project, the log-file is placed in the directory 'logs/XXX', where
	 * XXX is a unique name composed of the project name and the time when the current
	 * simulation started. Use this feature to not overwrite log-files in subsequent runs.
	 * <p>
	 * If append is set to true, the corresponding log-file is placed in the
	 * directory 'logs', ignoring the logToTimeDirectory flag of the configuration 
	 * file.
	 * If the specified file already exists, this logger object appends to it, 
	 * otherwise, a new file is created.
	 * @param logFileName The file name that identifies the logger object to be returned. 
	 * This name may also contain sub-directories, in which the corresponding log file is
	 * placed.
	 * @param append True to append to an already exising log file. If the file does 
	 * not yet exist, a new file is created.
	 * @return A logging object for the given log file name.  
	 */
	public static Logging getLogger(String logFileName, boolean append) {
			if(loggers.containsKey(logFileName)) {
				return loggers.get(logFileName);
			} 
			Logging l = new Logging(logFileName, append);
				loggers.put(logFileName, l);
				return l;
			
	}
	
	/**
	 * Adds a log-message to the log file, if the logFlag is set.
	 * @param logFlag Flag to enable/disable ths log-message
	 * @param txt
	 */
	public void log(boolean logFlag, String txt) {
		if(logFlag) {
			out.print(txt);
			if(properties.eagerLogFlush) {
				out.flush();
			}
		}
	}

	/**
	 * Adds a log-message to the log file.
	 * @param txt The text to log.
	 */
	public void log(String txt) {
		log(properties.logOn,txt);
	}
	
	/**
	 * Adds a log-message with line-break to the log file, if the logFlag 
	 * is set.
	 * @param logFlag Flag to enable/disable ths log-message
	 * @param txt The log message to be printed.
	 */
	public void logln(boolean logFlag, String txt) {
		log(logFlag,txt+"\n");
	}

	/**
	 * Adds a log-message with line-break to the log file.
	 * @param txt The log message to be printed.
	 */
	public void logln(String txt) {
		log(properties.logOn,txt+"\n");
	}
	


	/** 
	 * Returns the print stream where this logger logs to.
	 * @return The print stream where this logger logs to.
	 */
	public PrintStream getOutputStream() {
		return out;
	}
	
	/**
	 * @return The time-prefix used for the directories when
	 * logToTimeDirectory in the config file is enabled.
	 */
	public String getTimePrefix() {
		return timePrefix;
	}

	/**
	 * Returns a string representation of the code position
	 * of the caller of this method.
	 * <p>
	 * The method returns a string consisting of the form
	 * className.methodName:lineNumber
	 * where className is the fully qualified class name of the
	 * class in which the code of the caller of this method is located, 
	 * methodName is the name of the method in which the
	 * code of the caller of this method is located,
	 * and lineNumber indicates the line number in the source-file
	 * at which this method is being called.
	 * @return A human friendly description of the position in code
	 * of the calling method. 
	 */
	public static String getCodePosition() {
		return getCodePosition(1);
	}
	
	/**
	 * Returns a string representation of the code position
	 * of the caller of this method or one of its parents.
	 * <p>
	 * To obtain the string representation of the caller of this 
	 * method, set offset to 0. This is equivalent to {@link #getCodePosition()}.
	 * To obtain the code position where this caller's method is being
	 * called, set offset to 1. 
	 * Generally, to get the code position of the method call n steps back
	 * in the calling sequence, set offset to n.
	 * @param offset The offset in the calling sequence to the method
	 * call that triggered the caller of this method to execute.
	 * @return A human friendly description of the position in code
	 * of the calling method or one of its parents, the empty string
	 * if offset does not specify an existing method in the calling 
	 * sequence.
	 * @see #getCodePosition()
	 */
	public static String getCodePosition(int offset) {
		String result = "<cannot determine code position>";
		Exception e = new Exception(); // a dummy exception to get the stack trace
		StackTraceElement trace[] = e.getStackTrace();
		offset++; // zero-based array 
		if(trace.length > offset && offset > 0) {
			result = trace[offset].getClassName() + "." + trace[offset].getMethodName() + ":" + trace[offset].getLineNumber();
		}
		return result;
	}	

	/**
	 * @return Returns a string representing the current time in the form
	 * Day.Month.Year-Hour:Minutes:Seconds.MilliSeconds
	 */
	public static String getTimeStamp() {
		SimpleDateFormat df = new SimpleDateFormat("dd.MM.yyyy-HH:mm:ss.SSS");
		return df.format(new Date());
	}
	
	//-----------------------------------------------------------------------------------
	//-----------------------------------------------------------------------------------
	// Framework specific methods and member variables
	// => You should not need to modify/overwrite/call/use any of these members or methods
	//-----------------------------------------------------------------------------------
	//-----------------------------------------------------------------------------------

	private static Logging instance = null;
	private static HashMap<String, Logging> loggers = new HashMap<String, Logging>();
	private PrintStream out;
	private static String timePrefix; // the time when the simulation started - can be prefixed to the log-files to distringish different rounds. 

	/**
	 * <b>This member is framework internal and should not be used by the project developer.</b> 
	 * Creates a directory if it does not already exist.
	 * @param dir
	 */
	private void createDir(String dir) {
		File f = new File(dir);
		if(f.exists() && !f.isDirectory()) {
			System.out.println("Cannot create folder '" + dir + "'. There is a file called the same name.");
		} else if(!f.exists()){
			try {
				if(!f.mkdirs()) {
					System.out.println("Could not generate all of the directories '" + dir + "'.");	
				}
			} catch(SecurityException e) {
				System.out.println("Cannot create folder '" + dir + "':\n" + e);
			}
		}
	}
	
	/**
	 * <b>This member is framework internal and should not be used by the project developer.</b>   
	 * Private constructor - this is a singleton implementation
	 * @param aFileName The file name of the file this logger should print to.
	 * @param append Set to true if this logger should append to an existing file. 
	 * If append is set to true, the file is never placed in the time-directory of
	 * the current simulation.
	 */
	private Logging(String aFileName, boolean append) {
		try {
			String dir = properties.logFileDirectory;
			if(dir != "") {
				createDir(dir);
				dir += "/";
			}
			
			if(!append) {
				if(properties.logToTimeDirectory) {
					dir = dir + properties.projectName + "_" + timePrefix;
					createDir(dir);
					dir = dir + "/";
				}
			} 
			int index = aFileName.lastIndexOf('/');
			if(index > 0) {
				String path = aFileName.substring(0, index);
				createDir(dir + path);
			}

			if(append) {
				out = new PrintStream(new FileOutputStream(dir + aFileName, true));	
			} else {
				out = new PrintStream(dir + aFileName);
			}
		} catch(FileNotFoundException e) {
			System.out.println("Could not open the logfile "+aFileName);
		}
	}

	/**
	 * <b>This member is framework internal and should not be used by the project developer.</b> 
	 * Private constructor - this is a singleton implementation 
	 * @param aStream The stream this logger should print to.
	 */
	private Logging(PrintStream aStream) {
		out = aStream;
	}
}
